import {
    AbstractFactory,
    Container,
    DependencyRequirement,
    PropertyDependency,
    isInstantiable,
    DEPENDENCY_KEYS_METADATA_KEY,
    DEPENDENCY_KEYS_PROPERTY_METADATA_KEY
} from "@extollo/di"
import {Collection, ErrorWithContext} from "@extollo/util"
import {MemorySession} from "./MemorySession";
import {Session} from "./Session";
import {Logging} from "../../service/Logging";
import {Config} from "../../service/Config";

export class SessionFactory extends AbstractFactory {
    protected readonly logging: Logging
    protected readonly config: Config

    private static loggedMemorySessionWarningOnce = false

    constructor() {
        super({})
        this.logging = Container.getContainer().make<Logging>(Logging)
        this.config = Container.getContainer().make<Config>(Config)
    }

    produce(dependencies: any[], parameters: any[]): Session {
        return new (this.getSessionClass())
    }

    match(something: any) {
        return something === Session
    }

    getDependencyKeys(): Collection<DependencyRequirement> {
        const meta = Reflect.getMetadata(DEPENDENCY_KEYS_METADATA_KEY, this.getSessionClass())
        if ( meta ) return meta
        return new Collection<DependencyRequirement>()
    }

    getInjectedProperties(): Collection<PropertyDependency> {
        const meta = new Collection<PropertyDependency>()
        let currentToken = this.getSessionClass()

        do {
            const loadedMeta = Reflect.getMetadata(DEPENDENCY_KEYS_PROPERTY_METADATA_KEY, currentToken)
            if ( loadedMeta ) meta.concat(loadedMeta)
            currentToken = Object.getPrototypeOf(currentToken)
        } while (Object.getPrototypeOf(currentToken) !== Function.prototype && Object.getPrototypeOf(currentToken) !== Object.prototype)

        return meta
    }

    protected getSessionClass() {
        const SessionClass = this.config.get('server.session.driver', MemorySession)
        if ( SessionClass === MemorySession && !SessionFactory.loggedMemorySessionWarningOnce ) {
            this.logging.warn(`You are using the default memory-based session driver. It is recommended you configure a persistent session driver instead.`)
            SessionFactory.loggedMemorySessionWarningOnce = true
        }

        if ( !isInstantiable(SessionClass) || !(SessionClass.prototype instanceof Session) ) {
            const e = new ErrorWithContext('Provided session class does not extend from @extollo/lib.Session');
            e.context = {
                config_key: 'server.session.driver',
                class: SessionClass.toString(),
            }
        }

        return SessionClass
    }
}
